from typing import Any
from uuid import UUID, uuid4

from pydantic import Field

from schorle.component import Component
from schorle.element import body, div, footer, head, link, meta, script, title
from schorle.page import Page
from schorle.tags import HTMLTag
from schorle.text import text
from schorle.theme import Theme


class Document(Component):
    tag: HTMLTag = HTMLTag.HTML
    title: str = "Schorle"
    csrf_token: UUID = Field(default_factory=uuid4)
    theme: Theme = Theme.DARK
    page: Page | None = None
    with_dev_meta: bool = False
    extra_assets: list | None = None
    lang: str = "en"
    with_tailwind: bool = True
    with_daisyui: bool = True
    with_htmx: bool = True
    daisyui_version: str = "4.7.2"

    def model_post_init(self, __context: Any) -> None:
        self.attributes["lang"] = self.lang
        super().model_post_init(__context)

    def render(self):
        with head():
            meta(charset="utf-8")
            meta(name="viewport", content="width=device-width, initial-scale=1.0")
            meta(name="schorle-csrf-token", content=str(self.csrf_token))
            if self.with_dev_meta:
                meta(name="schorle-dev", content="true")
            link(href="/favicon.svg", rel="icon", type="image/svg+xml")
            if self.with_tailwind:
                script(src="https://cdn.tailwindcss.com")
            if self.with_daisyui:
                link(
                    href=f"https://cdn.jsdelivr.net/npm/daisyui@{self.daisyui_version}/dist/full.min.css",
                    rel="stylesheet",
                    type="text/css",
                )
            if self.with_htmx:
                script(src="https://unpkg.com/htmx.org@1.9.10", crossorigin="anonymous")
                script(src="https://unpkg.com/htmx.org@1.9.10/dist/ext/ws.js")
                script(src="https://unpkg.com/idiomorph@0.3.0")
                script(src="https://unpkg.com/htmx.org/dist/ext/event-header.js")

            script(src="/_schorle/assets/bundle.js", crossorigin="anonymous")
            if self.extra_assets:
                for asset in self.extra_assets:
                    asset()

            with title():
                text(self.title)
        with body():
            with div(element_id="schorle-morph-wrapper", **{"hx-ext": "morph"}):
                with div(
                    element_id="schorle-event-handler",
                    **{
                        "hx-ext": "ws,event-header",
                        "ws-connect": "/_schorle/events",
                    },
                ):
                    if self.page:
                        with self.page:
                            self.page()
            if self.with_dev_meta:
                footer(element_id="schorle-footer")



import asyncio
from asyncio import Protocol
from collections.abc import Iterable
from dataclasses import dataclass
from functools import wraps
from inspect import ismethod
from types import MethodType
from typing import Callable

from pydantic import BaseModel


class Effector:
    def __init__(self, bounded_method: MethodType):
        self.pre_actions: list[Callable] = []
        self.bounded_method = bounded_method
        self.subscribers: list[Callable] = []

    async def __call__(self, *args, **kwargs):
        _tasks = [asyncio.create_task(_pre_action()) for _pre_action in self.pre_actions]
        await asyncio.gather(*_tasks)

        if asyncio.iscoroutinefunction(self.bounded_method):
            await self.bounded_method(*args[1:], **kwargs)
        else:
            self.bounded_method(*args[1:], **kwargs)

        for subscriber in self.subscribers:
            subscriber()

    def subscribe(self, callback):
        self.subscribers.append(callback)

    def prepend(self, _pre_action):
        self.pre_actions.append(_pre_action)


class EffectorProtocol(Protocol):
    def subscribe(self, callback, *, trigger: bool = True): ...

    async def __call__(self, *args, **kwargs): ...

    def prepend(self, _pre_action): ...


def create_emitter(func: MethodType) -> EffectorProtocol:
    _emitter_instance = Effector(func)

    async def _wrapper(*args, **kwargs):
        await _emitter_instance(*args, **kwargs)

    wrapper: EffectorProtocol = wraps(func)(_wrapper)  # type: ignore[assignment]
    wrapper.subscribe = _emitter_instance.subscribe  # type: ignore[method-assign, assignment]
    wrapper.prepend = _emitter_instance.prepend  # type: ignore[method-assign, assignment]
    return wrapper


def effector(func: Callable) -> EffectorProtocol:
    if not asyncio.iscoroutinefunction(func):
        msg = f"Effector must be a coroutine function. {func.__name__} is not a coroutine function"
        raise ValueError(msg)
    func.is_emitter = True  # type: ignore[attr-defined]
    return func  # type: ignore[return-value]


@dataclass
class EffectorInfo:
    method: MethodType
    attr: str


def effector_listing(_object: object) -> Iterable[EffectorInfo]:
    if isinstance(_object, BaseModel) and not _object.model_config.get("extra", None) == "allow":
        raise ValueError("Pydantic models must be declared with extra='allow' to use inject_emitters")
    for attr in dir(_object):
        if attr not in ["__fields__", "__fields_set__", "__signature__"] and (
            attr not in _object.model_computed_fields.keys() if isinstance(_object, BaseModel) else True
        ):
            if callable(getattr(_object, attr)) and ismethod(getattr(_object, attr)):
                _method: MethodType = getattr(_object, attr)
                if hasattr(_method.__func__, "is_emitter"):
                    yield EffectorInfo(_method, attr)


def inject_effectors(_object: object):
    for effector_info in effector_listing(_object):
        emitter_func = create_emitter(effector_info.method)
        emitter_method = MethodType(emitter_func, _object)
        setattr(_object, effector_info.attr, emitter_method)



from pydantic import Field

from schorle.classes import Classes
from schorle.component import Component
from schorle.effector import effector
from schorle.element import button, div, span
from schorle.page import Page
from schorle.state import ReactiveModel
from schorle.text import text
from schorle.types import LXMLElement
from schorle.utils import render_in_context


class Sample(ReactiveModel):
    count: int = 0

    @effector
    async def increment(self):
        self.count += 1


class Another(Component):
    def render(self):
        with span():
            text("Something")


class View(Component):
    counter: Sample

    def render(self):
        with button(classes=Classes("btn btn-primary")):
            text(f"Clicked {self.counter.count} times")
        Another()

    def __init__(self, **data):
        super().__init__(**data)
        self.bind(self.counter)


class ViewPage(Page):
    counter: Sample = Field(default_factory=Sample)

    def render(self):
        with div(classes=Classes("flex justify-center items-center h-screen")):
            View(counter=self.counter, element_id="v1")
        View(counter=self.counter)


def test_reactive():
    vp = ViewPage()

    with vp:
        r1 = render_in_context(vp)
        assert isinstance(r1, LXMLElement)



import base64
from pathlib import Path

from schorle.classes import Classes
from schorle.element import img
from schorle.on import On


def Image(  # noqa: N802
    src: str | Path,
    alt: str,
    media_type: str | None = None,
    element_id: str | None = None,
    classes: Classes | None = None,
    style: dict[str, str] | None = None,
    on: list[On] | On | None = None,
    **attributes,
) -> None:
    if isinstance(src, Path):
        if not media_type:
            raise ValueError("media_type must be provided when src is a Path object")

        _src = f"data:{media_type};base64,{base64.b64encode(src.read_bytes()).decode()}"
    else:
        _src = src

    attributes["src"] = _src
    attributes["alt"] = alt

    return img(element_id=element_id, classes=classes, style=style, on=on, **attributes)



import contextvars
from typing import Any

from lxml import etree

from schorle.types import LXMLElement


class RenderController:
    def __init__(self, page: Any | None = None):
        self._root: LXMLElement = etree.Element("root")
        self.previous: LXMLElement = self._root
        self.current: LXMLElement = self._root
        self.page: Any | None = page
        self.component_root: LXMLElement | None = None

    def get_root(self) -> LXMLElement:
        return self._root


class RenderControllerMixin:
    @property
    def controller(self) -> RenderController:
        return RENDER_CONTROLLER.get()


RENDER_CONTROLLER: contextvars.ContextVar[RenderController] = contextvars.ContextVar(
    "render_controller", default=RenderController()
)



import sys
from enum import Enum
from typing import Any

from schorle.render_controller import RENDER_CONTROLLER, RenderController
from schorle.types import LXMLElement


class RunningMode(str, Enum):
    DEV = "dev"
    PRODUCTION = "production"


def get_running_mode() -> RunningMode:
    # if uvicorn and --reload are in sys.argv, we are running in uvicorn dev mode
    _joined_argv = " ".join(sys.argv)
    if "uvicorn" in _joined_argv and "--reload" in _joined_argv:
        return RunningMode.DEV
    else:
        return RunningMode.PRODUCTION


def render_in_context(component: Any, page: Any | None = None) -> LXMLElement:
    token = RENDER_CONTROLLER.set(RenderController(page=page))
    try:
        component()
        return RENDER_CONTROLLER.get().get_root().getchildren()[0]  # type: ignore[attr-defined]
    finally:
        RENDER_CONTROLLER.reset(token)



from enum import Enum


class Theme(str, Enum):
    LIGHT = "light"
    DARK = "dark"
    CUPCAKE = "cupcake"
    BUMBLEBEE = "bumblebee"
    EMERALD = "emerald"
    CORPORATE = "corporate"
    SYNTHWAVE = "synthwave"
    RETRO = "retro"
    CYBERPUNK = "cyberpunk"
    VALENTINE = "valentine"
    HALLOWEEN = "halloween"
    GARDEN = "garden"
    FOREST = "forest"
    AQUA = "aqua"
    LOFI = "lofi"
    PASTEL = "pastel"
    FANTASY = "fantasy"
    WIREFRAME = "wireframe"
    BLACK = "black"
    LUXURY = "luxury"
    DRACULA = "dracula"
    CMYK = "cmyk"
    AUTUMN = "autumn"
    BUSINESS = "business"
    ACID = "acid"
    LEMONADE = "lemonade"
    NIGHT = "night"
    COFFEE = "coffee"
    WINTER = "winter"
    DIM = "dim"
    NORD = "nord"
    SUNSET = "sunset"



from lxml import etree

from schorle.component import Component
from schorle.element import button, div
from schorle.page import Page
from schorle.tags import HTMLTag
from schorle.text import text
from schorle.utils import render_in_context


class Button(Component):
    tag: HTMLTag = HTMLTag.BUTTON

    def render(self):
        text("Hey!")


class SamplePage(Page):
    def render(self):
        with div():
            with button():
                text("Hello, World!")


def test_context_rendering():
    page = SamplePage()
    rendered = render_in_context(page)
    assert isinstance(rendered, etree._Element)



from typing import Annotated

from pydantic import BaseModel, ConfigDict, Field


class HtmxHeaders(BaseModel):
    request: Annotated[bool, Field(alias="HX-Request")]
    trigger_element_id: Annotated[str, Field(alias="HX-Trigger")]
    trigger_type: Annotated[str | None, Field(None, alias="HX-Trigger-Type")]
    trigger_name: Annotated[str | None, Field(None, alias="HX-Trigger-Name")]


class HtmxMessage(BaseModel):
    model_config = ConfigDict(extra="allow")
    headers: Annotated[HtmxHeaders, Field(alias="HEADERS")]



from functools import partial
from pathlib import Path

from pydantic import Field
from starlette.responses import FileResponse, JSONResponse

from schorle.app import Schorle
from schorle.classes import Classes
from schorle.component import Component
from schorle.element import div, link, p
from schorle.icon import Icon
from schorle.img import Image
from schorle.page import Page
from schorle.tags import HTMLTag
from schorle.text import text

app = Schorle(
    extra_assets=[
        partial(link, href="https://fonts.googleapis.com/css?family=Space+Mono", rel="stylesheet", type="text/css"),
    ]
)


class LinkWithIcon(Component):
    href: str
    icon_name: str
    text: str
    tag: HTMLTag = HTMLTag.A
    classes: Classes = Classes("btn btn-primary font-normal w-42")

    def model_post_init(self, __context):
        self.attributes["href"] = self.href
        super().model_post_init(__context)

    def render(self):
        text(self.text)
        Icon(name=self.icon_name)


LINKS = [
    LinkWithIcon(href="https://github.com/renardeinside/schorle", icon_name="github", text="GitHub"),
    LinkWithIcon(href="https://github.com/renardeinside/schorle/tree/main/examples", icon_name="code", text="Examples"),
    LinkWithIcon(
        href="https://medium.com/@polarpersonal/schorle-testing-the-waters-with-a-python-server-driven-ui-kit-053f85ee6574",
        icon_name="book-marked",
        text="Concepts",
    ),
]


@app.backend.get("/logo", response_class=FileResponse)
def logo():
    return FileResponse(path=Path(__file__).parent.parent / Path("raw/with_text.svg"), media_type="image/svg+xml")


class LandingPage(Page):
    style: dict[str, str] = Field(default_factory=lambda: {"font-family": "Space Mono"})

    def render(self):
        with div(classes=Classes("flex flex-col justify-center items-center h-screen")):
            with div(classes=Classes("max-w-screen-md flex flex-col justify-center items-center space-y-4")):
                Image(
                    src="/logo",
                    alt="Schorle logo",
                    media_type="image/svg+xml",
                    classes=Classes("w-3/4"),
                )
            with p(classes=Classes("text-2xl m-4 p-4 w-5/6 text-center")):
                text("Pythonic Server-Driven UI Kit for building modern apps.")

            with div(classes=Classes("flex flex-col space-y-4 space-x-0 md:flex-row md:space-x-4 md:space-y-0")):
                for _link in LINKS:
                    _link()


@app.get("/")
def landing_page():
    return LandingPage()


@app.backend.get("/health", response_class=JSONResponse)
def health():
    return {"status": "ok"}



from typing import Any

from pydantic import Field

from schorle.classes import Classes
from schorle.component import Component
from schorle.tags import HTMLTag


class TextInput(Component):
    tag: HTMLTag = HTMLTag.INPUT
    base_classes: Classes = Field(default_factory=lambda: Classes("input", "form-control"))
    placeholder: str = ""
    value: str = ""
    name: str = Field(default="default")

    def model_post_init(self, __context: Any):
        self.classes.append(self.base_classes)
        self.attributes["placeholder"] = self.placeholder
        self.attributes["value"] = self.value
        self.attributes["hx-include"] = "this"
        self.attributes["name"] = self.name
        super().model_post_init(__context)

    def render(self):
        pass



from __future__ import annotations

from typing import Union

from pydantic import BaseModel, PrivateAttr

RawClassesPayload = Union[str, list[str], tuple[str, ...], "Classes", None]


def parse_args(*args: RawClassesPayload) -> list[str]:
    container = []
    for arg in args:
        if isinstance(arg, str):
            container.append(arg)
        elif isinstance(arg, (list, tuple)):
            container.extend(arg)
        elif isinstance(arg, Classes):
            container.extend(arg._value)
        elif arg is None:
            pass
        else:
            msg = f"Invalid type: {type(arg)} of {arg} for Classes"
            raise TypeError(msg)
    return container


class Classes(BaseModel):
    _value: list[str] = PrivateAttr(default_factory=list)

    def __init__(self, *args: RawClassesPayload):
        super().__init__()
        self._value = parse_args(*args)

    def append(self, *args: RawClassesPayload):
        new_value = self._value + parse_args(*args)
        self._value = new_value
        return self

    def remove(self, *args: RawClassesPayload):
        new_value = [x for x in self._value if x not in parse_args(*args)]
        self._value = new_value
        return self

    def toggle(self, class_name: str):
        if class_name in self._value:
            self.remove(class_name)
        else:
            self.append(class_name)
        return self

    def render(self) -> str:
        return "" if not self._value else " ".join(set(self._value)).strip()



from pydantic import BaseModel


class Attribute(BaseModel):
    alias: str
    value: str

    def __init__(self, alias: str, value: str):
        super().__init__(**{"alias": alias, "value": value})


class Id(Attribute):
    alias: str = "id"
    value: str

    def __init__(self, value: str):
        super().__init__(**{"alias": "id", "value": value})



import requests
from lxml import etree

from schorle.component import Component


class Icon(Component):
    name: str

    def render(self):
        icon_url = f"https://raw.githubusercontent.com/lucide-icons/lucide/main/icons/{self.name}.svg"
        response = requests.get(icon_url, timeout=5)
        response.raise_for_status()
        parser = etree.XMLParser(ns_clean=True, resolve_entities=False, no_network=True)
        # we can disable security check here because we've set the parser to not resolve entities
        # read https://github.com/PyCQA/bandit/issues/767 for more information
        icon = etree.fromstring(response.text, parser)  # noqa: S320
        self.controller.current.append(icon)



import asyncio

from loguru import logger
from lxml import etree
from starlette.websockets import WebSocket

from schorle.component import Component
from schorle.page import Page
from schorle.types import LXMLElement
from schorle.utils import render_in_context


class PageEmitter:
    def __init__(self, page: Page) -> None:
        self._page = page

    async def emit(self, ws: WebSocket) -> None:
        while True:
            try:
                await asyncio.sleep(0.0001)
                renderable = await self._page._render_queue.get()
                if isinstance(renderable, LXMLElement):
                    _html = etree.tostring(renderable, pretty_print=True).decode()
                elif isinstance(renderable, Component):
                    with self._page:
                        rendered = render_in_context(renderable, self._page)
                        _html = etree.tostring(rendered, pretty_print=True).decode()
                await ws.send_text(_html)
            except Exception as e:
                logger.error(f"Error while emitting: {e}")
                break



from schorle.classes import Classes
from schorle.element import button
from schorle.on import On
from schorle.suspense import Suspense


def Button(  # noqa: N802
    element_id: str | None = None,
    modifier: str | None = None,
    disabled: bool | None = None,
    classes: Classes | None = None,
    style: dict[str, str] | None = None,
    on: list[On] | On | None = None,
    suspense: Suspense | None = None,
    **attributes,
) -> None:
    _classes = Classes("btn")
    if modifier:
        _classes.append(f"btn-{modifier}")
    if disabled:
        _classes.append("btn-disabled")
    if classes:
        _classes.append(classes)

    return button(element_id=element_id, classes=_classes, style=style, on=on, suspense=suspense, **attributes)



from __future__ import annotations

from schorle.app import Schorle
from schorle.button import Button
from schorle.classes import Classes
from schorle.effector import effector
from schorle.element import div, p
from schorle.on import On
from schorle.page import Page
from schorle.state import ReactiveModel
from schorle.text import text

app = Schorle()


class Counter(ReactiveModel):
    value: int = 0

    @effector
    async def increment(self):
        self.value += 1

    @effector
    async def decrement(self):
        self.value -= 1


class PageWithButton(Page):
    counter: Counter = Counter.factory()

    def render(self):
        with div(classes=Classes("flex flex-col justify-center items-center h-screen")):
            with div(classes=Classes("flex flex-row justify-center items-center space-x-4")):
                with Button(on=On("click", self.counter.increment), modifier="primary"):
                    text("Increment")
                with Button(
                    on=On("click", self.counter.decrement),
                    modifier="secondary",
                    disabled=self.counter.value <= 0,
                ):
                    text("Decrement")
            with p(classes=Classes("text-2xl")):
                text(f"Counter: {self.counter.value}")

    def initialize(self):
        self.bind(self.counter)


@app.get("/")
def get_page():
    return PageWithButton()



import asyncio
from functools import partial
from random import random

from pydantic import Field

from schorle.app import Schorle
from schorle.button import Button
from schorle.classes import Classes
from schorle.component import Component
from schorle.effector import effector
from schorle.element import div, p
from schorle.inputs import TextInput
from schorle.loading import Loading
from schorle.on import On
from schorle.page import Page
from schorle.state import ReactiveModel
from schorle.suspense import Suspense
from schorle.text import text

app = Schorle()


class State(ReactiveModel):
    current: str = ""
    todos: list[str] = Field(default_factory=lambda: ["Buy milk", "Do laundry"])

    @effector
    async def set_current(self, value):
        self.current = value

    @effector
    async def add_todo(self):
        await asyncio.sleep(random() * 2)  # Simulate a slow network request
        self.todos.append(self.current)
        await self.set_current("")

    @effector
    async def remove(self, todo):
        await asyncio.sleep(random() * 2)  # Simulate a slow network request
        self.todos.remove(todo)


class InputSection(Component):
    state: State

    def render(self):
        with div(classes=Classes("flex flex-row w-3/4 justify-center items-center space-x-4")):
            TextInput(
                value=self.state.current,
                placeholder="Add a todo",
                classes=Classes("input-bordered"),
                on=On("change", self.state.set_current),
            )
            with Button(on=On("click", self.state.add_todo), modifier="primary"):
                text("Add")

    def initialize(self):
        self.bind(self.state)


class TodoView(Component):
    state: State
    classes: Classes = Classes("w-full flex flex-col items-center")

    def render(self):
        with div(
            classes=Classes("flex flex-col items-center space-y-2 w-3/4 max-w-md m-4 min-h-96"),
            suspense=Suspense(on=self.state, fallback=Loading()),
        ):
            with p(classes=Classes("text-xl")):
                _text = "No todos yet." if not self.state.todos else "Your todos:"
                text(_text)

            for todo in self.state.todos:
                with div(classes=Classes("flex flex-row items-center justify-between w-full")):
                    with p():
                        text(todo)
                    with Button(
                        on=On("click", partial(self.state.remove, todo)),
                        modifier="error",
                    ):
                        text("Delete")

    def initialize(self):
        self.bind(self.state)


class TodoListPage(Page):
    state: State = State.factory()
    classes: Classes = Classes("m-4")

    def render(self):
        with div(classes=Classes("flex flex-col items-center h-screen")):
            InputSection(state=self.state)
            TodoView(state=self.state)


@app.get("/")
def index():
    return TodoListPage()



from lxml import etree

from schorle.classes import Classes
from schorle.document import Document
from schorle.element import button, div
from schorle.page import Page
from schorle.text import text
from schorle.utils import render_in_context


def test_doc():
    class SamplePage(Page):
        def render(self):
            with div(classes=Classes("flex flex-col justify-center items-center h-screen")):
                with div(classes=Classes("flex flex-row justify-center items-center")):
                    with button(classes=Classes("btn btn-primary")):
                        text("Increment")
                    with button(classes=Classes("btn btn-secondary")):
                        text("Decrement")
                with div(classes=Classes("text-2xl")):
                    text("Counter something")

    doc = Document(title="Test Document", page=SamplePage())
    root = render_in_context(doc)
    result = etree.tostring(root, pretty_print=True).decode()
    assert isinstance(result, str)



from schorle.classes import Classes
from schorle.component import Component
from schorle.element import span
from schorle.text import text


class Loading(Component):
    lazy_append: bool = True

    def render(self):
        with span(classes=Classes("loading loading-lg loading-infinity")):
            text("")



from copy import deepcopy
from typing import Any

from loguru import logger

from schorle.effector import EffectorProtocol
from schorle.render_controller import RenderControllerMixin
from schorle.state import ReactiveModel
from schorle.types import LXMLElement
from schorle.utils import render_in_context


class Suspense(RenderControllerMixin):
    def __init__(self, on: EffectorProtocol | ReactiveModel, fallback: Any):
        self.on = on
        self._parent: LXMLElement | None = None
        self._page_ref = self.controller.page
        self._fallback = render_in_context(fallback, None)

        async def _pre_action():
            if self._parent is None:
                logger.warning("Parent is not set")
                return
            else:
                _copy = deepcopy(self._parent)
                required_attrs = ["id", "classes", "style", "hx-swap-oob"]
                _saved = {k: _copy.get(k) for k in required_attrs}
                _copy.clear()
                for k, v in _saved.items():
                    if v is not None:
                        _copy.set(k, v)
                _copy.append(self._fallback)
                self._page_ref.append_to_queue(_copy)

        if isinstance(on, ReactiveModel):
            for effector_info in on.get_effectors():
                effector_info.method.prepend(_pre_action)
        elif isinstance(on, EffectorProtocol):
            self.on.prepend(_pre_action)



from pydantic import BaseModel, Field

from schorle.effector import effector_listing, inject_effectors


class EffectorMixin:
    def get_effectors(self):
        return effector_listing(self)


class ReactiveModel(BaseModel, EffectorMixin, extra="allow"):
    def __init__(self, **data):
        super().__init__(**data)
        inject_effectors(self)

    @classmethod
    def factory(cls, **data):
        return Field(default_factory=lambda: cls(**data))


class ReactiveState(EffectorMixin):
    def __init__(self):
        inject_effectors(self)



from abc import abstractmethod
from asyncio import Queue
from typing import Any

from pydantic import Field, PrivateAttr

from schorle.component import Component
from schorle.tags import HTMLTag


class Page(Component):
    tag: HTMLTag = HTMLTag.DIV
    element_id: str = "schorle-page"
    reactives: dict[str, Any] = Field(default_factory=dict)
    suspense: dict[str, Any] = Field(default_factory=dict)
    _render_queue: Queue[Component] = PrivateAttr(default_factory=Queue)

    def append_to_queue(self, component: Component):
        self._render_queue.put_nowait(component)

    @abstractmethod
    def render(self):
        pass

    def __enter__(self):
        self.controller.page = self
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.controller.page = None
        pass



import asyncio
import pkgutil
from asyncio import Task, iscoroutinefunction
from collections.abc import Callable
from functools import partial
from importlib.resources import files
from pathlib import Path

from fastapi import FastAPI
from loguru import logger
from lxml import etree
from starlette.endpoints import WebSocketEndpoint
from starlette.responses import FileResponse, HTMLResponse, PlainTextResponse
from starlette.types import Receive, Scope, Send
from starlette.websockets import WebSocket

from schorle.document import Document
from schorle.emitter import PageEmitter
from schorle.models import HtmxMessage
from schorle.page import Page
from schorle.theme import Theme
from schorle.utils import RunningMode, get_running_mode, render_in_context


def favicon() -> FileResponse:
    package_path = files("schorle")
    favicon_path = Path(str(package_path)) / "assets" / "favicon.svg"
    return FileResponse(favicon_path, media_type="image/svg+xml")


def assets(file_name: str) -> PlainTextResponse:
    _bundle = pkgutil.get_data("schorle", f"assets/{file_name}")
    if _bundle:
        return PlainTextResponse(_bundle.decode("utf-8"), status_code=200)
    else:
        return PlainTextResponse(f"File not found: {file_name}", status_code=404)


class Schorle:
    def __init__(self, theme: Theme = Theme.DARK, lang: str = "en", extra_assets: list | None = None) -> None:
        self._pages: dict[str, Page] = {}
        self.backend = FastAPI()
        self.backend.get("/_schorle/assets/{file_name:path}")(assets)
        self.backend.add_websocket_route("/_schorle/events", partial(EventsEndpoint, pages=self._pages))
        self.backend.get("/favicon.svg", response_class=FileResponse)(favicon)
        self.theme: Theme = theme
        self.extra_assets = extra_assets
        self.lang = lang

    def get(self, path: str):
        def decorator(func: Callable[..., Page]):
            self.backend.get(path, response_class=HTMLResponse)(partial(self.render_to_response, func))
            return func

        return decorator

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        """
        This method is called by uvicorn when the server is started.
        """
        await self.backend(scope=scope, receive=receive, send=send)

    async def render_to_response(self, page_provider: Callable[..., Page]) -> HTMLResponse:
        page = page_provider()
        logger.info(f"Rendering page: {page}...")

        if get_running_mode() == RunningMode.DEV:
            logger.info("Adding dev meta tags...")

        doc = Document(
            title="Schorle",
            page=page,
            theme=self.theme,
            with_dev_meta=get_running_mode() == RunningMode.DEV,
            extra_assets=self.extra_assets,
            lang=self.lang,
        )
        logger.debug(f"Rendering page: {page} with theme: {self.theme}...")

        lxml_element = render_in_context(doc)
        rendered = etree.tostring(lxml_element, pretty_print=True, doctype="<!DOCTYPE html>").decode("utf-8")
        response = HTMLResponse(rendered, status_code=200)

        logger.info(f"Adding page to cache with token: {doc.csrf_token}")

        self._pages[str(doc.csrf_token)] = page

        logger.debug("Page rendered.")

        return response


class EventsEndpoint(WebSocketEndpoint):
    encoding = "text"

    def __init__(self, scope, receive, send, pages: dict[str, Page]) -> None:
        super().__init__(scope, receive, send)
        self._page: Page | None = None
        self._pages: dict[str, Page] = pages
        self.page_emitter_task: Task | None = None

    async def on_connect(self, websocket: WebSocket) -> None:
        token = websocket.query_params.get("token")
        if not token:
            logger.error("No token provided.")
            await websocket.close()
            return
        page: Page | None = self._pages.get(token)
        if page:
            await websocket.accept()
            self._page = page
            self.page_emitter_task = asyncio.create_task(PageEmitter(page).emit(websocket))
            logger.info("Events connected.")

        elif not page and get_running_mode() == RunningMode.DEV:
            logger.info("Sending reload message to client...")
            await websocket.accept()
            await websocket.send_text("reload")  # standard schorle reload message
            await websocket.close()
            return
        else:
            logger.error(f"No page found for token: {token}")
            await websocket.close()
            return

    async def on_receive(self, ws: WebSocket, data: str) -> None:
        logger.warning(f"Events received message: {data}")
        message = HtmxMessage.model_validate_json(data)
        logger.debug(f"Events received message: {message}")
        if self._page:
            reactive = self._page.reactives.get(message.headers.trigger_element_id)
            if reactive:
                logger.debug(f"Events found reactive: {reactive}")
                _callback = reactive.get(message.headers.trigger_type)
                if _callback:
                    logger.debug(f"Events found callback: {_callback}")

                    if message.headers.trigger_name is not None:
                        _value = getattr(message, message.headers.trigger_name)
                        _cb = partial(_callback, _value)
                    else:
                        _cb = _callback

                    # TODO: catch exceptions in _cb
                    asyncio.ensure_future(_cb())  # noqa: RUF006

                else:
                    logger.error(f"Events no callback found for type: {message.headers.trigger_type}")
            else:
                logger.error(f"Events no reactive found for id: {message.headers.trigger_element_id}")

        else:
            logger.error("No page found, closing websocket...")
            await ws.close()

    async def on_disconnect(self, _: WebSocket, close_code: int) -> None:
        logger.info(f"Events disconnected with code: {close_code}")
        if self.page_emitter_task:
            self.page_emitter_task.cancel()
            logger.info("Events emitter task cancelled.")


def wrap_in_coroutine(func: Callable) -> Callable:
    if iscoroutinefunction(func):
        return func
    else:

        async def _wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return _wrapper



from __future__ import annotations

from pydantic import Field

from schorle.app import Schorle
from schorle.classes import Classes
from schorle.component import Component
from schorle.effector import effector
from schorle.element import button, div
from schorle.on import On
from schorle.page import Page
from schorle.state import ReactiveModel
from schorle.text import text

app = Schorle()


class Counter(ReactiveModel):
    count: int = 0

    @effector
    async def increment(self):
        self.count += 1

    @effector
    async def decrement(self):
        self.count -= 1


class Button(Component):
    counter: Counter

    def render(self):
        with button(on=On("click", self.counter.increment), classes=Classes("btn btn-primary")):
            text("Click me" if self.counter.count == 0 else f"Clicked {self.counter.count} times")

    def initialize(self):
        self.bind(self.counter)


class PageWithButton(Page):
    counter: Counter = Field(default_factory=Counter)

    def render(self):
        with div(classes=Classes("flex flex-col justify-center items-center h-screen")):
            Button(counter=self.counter)


@app.get("/")
def get_page():
    return PageWithButton()



from loguru import logger
from lxml import etree

from schorle.classes import Classes
from schorle.component import Component
from schorle.element import button, div, span
from schorle.page import Page
from schorle.text import text
from schorle.utils import render_in_context


class C(Component):
    def render(self):
        with button():
            with span(classes=Classes("htmx-indicator")):
                with span(classes=Classes("loading loading-infinity")):
                    pass

            with span():
                text("something")


class P(Page):
    def render(self):
        with div(classes=Classes("flex flex-col justify-center items-center h-screen")):
            C()


def test_nested():
    page = P()
    rendered = render_in_context(page, page=page)
    logger.debug(etree.tostring(rendered, pretty_print=True).decode())



import hashlib

from lxml import etree

from schorle.classes import Classes
from schorle.on import On
from schorle.render_controller import RenderControllerMixin
from schorle.suspense import Suspense
from schorle.tags import HTMLTag


class Element(RenderControllerMixin):
    def __init__(
        self,
        tag: HTMLTag,
        element_id: str | None = None,
        classes: Classes | None = None,
        style: dict[str, str] | None = None,
        on: list[On] | On | None = None,
        suspense: Suspense | None = None,
        **attributes,
    ):
        self.tag = tag.value
        self._element_id = element_id
        self._element = self.get_element()

        if attributes:
            for k, v in attributes.items():
                self._element.set(k, v)

        if style:
            self._element.set("style", ";".join([f"{k}: {v}" for k, v in style.items()]))
        if classes:
            _rendered = classes.render()
            if _rendered:
                self._element.set("class", _rendered)

        if element_id:
            self._element.set("id", self._element_id)

        if self.controller.page:
            self._element.set("hx-swap-oob", "morph")

        if self.controller.page:
            if on:
                on = [on] if isinstance(on, On) else on
                self._element.set("ws-send", "")
                _triggers = ",".join([o.trigger for o in on])
                self._element.set("hx-trigger", _triggers)

                if not self._element_id:
                    self._element_id = "sle-" + self._generate_hash("|".join(str(id(_on.callback)) for _on in on))[:8]
                    self._element.set("id", self._element_id)

                self.controller.page.reactives[self._element_id] = {_on.trigger: _on.callback for _on in on}

            if suspense:
                if not self._element_id:
                    self._element_id = "sle-" + self._generate_hash(f"suspense-{id(suspense.on)}")[:8]
                    self._element.set("id", self._element_id)
                suspense._parent = self._element

    @staticmethod
    def _generate_hash(string: str) -> str:
        return hashlib.sha256(string.encode("utf-8")).hexdigest()

    def get_element(self):
        elem = etree.SubElement(self.controller.current, self.tag)
        if self.tag in [HTMLTag.SCRIPT, HTMLTag.LINK]:
            elem.text = ""  # Prevents self-closing tags
        return elem

    def __enter__(self):
        self._pre_previous = self.controller.previous
        self.controller.previous = self.controller.current
        self.controller.current = self._element

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.controller.current = self.controller.previous
        self.controller.previous = self._pre_previous
        pass

    def add(self):
        with self:
            pass

    def __repr__(self):
        _id_str = f"id={self._element_id}" if self._element_id else ""
        return f"<{self.tag} {_id_str}/>"

    def __str__(self):
        return self.__repr__()


def element_function_factory(tag: HTMLTag):
    def func(
        element_id: str | None = None,
        classes: Classes | None = None,
        style: dict[str, str] | None = None,
        on: list[On] | On | None = None,
        suspense: Suspense | None = None,
        **attributes,
    ):
        return Element(tag, element_id, classes, style, on, suspense, **attributes)

    return func


div = element_function_factory(HTMLTag.DIV)
span = element_function_factory(HTMLTag.SPAN)
button = element_function_factory(HTMLTag.BUTTON)
input_ = element_function_factory(HTMLTag.INPUT)
a = element_function_factory(HTMLTag.A)
img = element_function_factory(HTMLTag.IMG)
script = element_function_factory(HTMLTag.SCRIPT)
link = element_function_factory(HTMLTag.LINK)
meta = element_function_factory(HTMLTag.META)
title = element_function_factory(HTMLTag.TITLE)
head = element_function_factory(HTMLTag.HEAD)
body = element_function_factory(HTMLTag.BODY)
html = element_function_factory(HTMLTag.HTML)
p = element_function_factory(HTMLTag.P)
h1 = element_function_factory(HTMLTag.H1)
h2 = element_function_factory(HTMLTag.H2)
h3 = element_function_factory(HTMLTag.H3)
h4 = element_function_factory(HTMLTag.H4)
h5 = element_function_factory(HTMLTag.H5)
h6 = element_function_factory(HTMLTag.H6)
ul = element_function_factory(HTMLTag.UL)
ol = element_function_factory(HTMLTag.OL)
li = element_function_factory(HTMLTag.LI)
table = element_function_factory(HTMLTag.TABLE)
tr = element_function_factory(HTMLTag.TR)
td = element_function_factory(HTMLTag.TD)
th = element_function_factory(HTMLTag.TH)
form = element_function_factory(HTMLTag.FORM)
footer = element_function_factory(HTMLTag.FOOTER)



from __future__ import annotations

import asyncio

from schorle.app import Schorle
from schorle.button import Button
from schorle.classes import Classes
from schorle.component import Component
from schorle.effector import effector
from schorle.element import div
from schorle.loading import Loading
from schorle.on import On
from schorle.page import Page
from schorle.state import ReactiveModel
from schorle.suspense import Suspense
from schorle.text import text

app = Schorle()


class Counter(ReactiveModel):
    count: int = 0

    @effector
    async def increment(self):
        await asyncio.sleep(1)
        self.count += 1


class StatefulButton(Component):
    counter: Counter = Counter.factory()
    modifier: str | None = None

    def render(self):
        with Button(
            on=On("click", self.counter.increment),
            suspense=Suspense(on=self.counter.increment, fallback=Loading()),
            classes=self.classes.append("w-48"),
            modifier=self.modifier,
        ):
            text("Click me" if self.counter.count == 0 else f"Clicked {self.counter.count} times")

    def initialize(self):
        self.bind(self.counter)


class PageWithButton(Page):

    def render(self):
        with div(classes=Classes("flex flex-col justify-center items-center h-screen")):
            with div(classes=Classes("space-y-2")):
                StatefulButton(modifier="primary")
                StatefulButton(modifier="secondary")


@app.get("/")
def get_page():
    return PageWithButton()



from abc import ABC, abstractmethod
from typing import Any
from uuid import uuid4

from pydantic import BaseModel, Field

from schorle.classes import Classes
from schorle.element import Element
from schorle.on import On
from schorle.render_controller import RenderControllerMixin
from schorle.state import ReactiveModel
from schorle.tags import HTMLTag


class Component(ABC, BaseModel, RenderControllerMixin):
    tag: HTMLTag = HTMLTag.DIV
    classes: Classes = Field(default_factory=Classes)
    style: dict[str, str] = Field(default_factory=dict)
    element_id: str | None = None
    attributes: dict[str, str] = Field(default_factory=dict)
    on: list[On] | On = Field(default_factory=list)
    _page_ref: Any | None = None
    lazy_append: bool = False

    def add(self):
        pre_previous = self.controller.previous
        pre_current = self.controller.current
        self._page_ref = self.controller.page

        with Element(self.tag, self.element_id, classes=self.classes, style=self.style, on=self.on, **self.attributes):
            self.render()

        self.controller.previous = pre_previous
        self.controller.current = pre_current

    def model_post_init(self, __context: Any) -> None:
        if not self.element_id and self.tag not in [HTMLTag.HTML, HTMLTag.BODY]:
            self.element_id = f"sle-{self.tag}-{str(uuid4())[:8]}"

        self.initialize()

        if self.controller.page and not self.lazy_append:
            self.add()

    def initialize(self):
        pass

    @abstractmethod
    def render(self):
        pass

    def __call__(self):
        self.add()

    def __repr__(self):
        return f"<{self.__class__.__name__}/>"

    def __str__(self):
        return self.__repr__()

    def bind(self, reactive: ReactiveModel):
        def _appender():
            self._page_ref.append_to_queue(self)

        for effector in reactive.get_effectors():
            effector.method.subscribe(_appender)
